"""

Tile server components

"""
import os
from fastapi import FastAPI, APIRouter, Response, Request, Depends
from fastapi.middleware.cors import CORSMiddleware
import large_image
import requests
import json
from typing import Annotated
from typing_extensions import Union

from copy import deepcopy
import numpy as np
import uvicorn

import asyncio

from shapely.geometry import box, shape

from fusion_tools.database.database import fusionDB 
from fusion_tools.utils.shapes import (
    load_annotations,
    histomics_to_geojson, 
    detect_image_overlay, 
    detect_geojson, 
    detect_histomics,
    structures_within_poly,
    extract_nested_prop
)


class Slide:
    """Local slide object with built-in methods for checking paths, loading annotations
    """
    def __init__(self,
                 image_filepath: Union[str,None] = None,
                 annotations: Union[str,list,dict,None] = None,
                 metadata: Union[dict,None] = None,
                 image_style: Union[dict,None] = None,
                 public: bool = False):
        
        self.image_filepath = image_filepath
        self.annotations = annotations
        self.metadata = metadata
        self.image_style = image_style
        self.public = public

        # Checking path exists
        assert(os.path.exists(self.image_filepath))
        
        # Checking image is readable by large_image
        if image_style is None:
            img_source = large_image.open(self.image_filepath)
        else:
            img_source = large_image.open(
                self.image_filepath,
                style = self.image_style
            )

        self.image_metadata = img_source.getMetadata()

        # Treating 3-frame images as RGB by default
        if 'frames' in self.image_metadata:
            if len(self.image_metadata['frames'])==3:
                new_tile_source = large_image.open(
                    self.image_filepath,
                    style = {
                        "bands": [
                            {
                                "framedelta": c_idx,
                                "palette": ["rgba(0,0,0,0)","rgba("+",".join(["255" if i==c_idx else "0" for i in range(3)]+["255"])+")"]
                            }
                            for c_idx in range(3)
                        ]
                    }
                )
                self.image_style = {
                    "bands": [
                        {
                            "framedelta": c_idx,
                            "palette": ["rgba(0,0,0,0)","rgba("+",".join(["255" if i==c_idx else "0" for i in range(3)]+["255"])+")"]
                        }
                        for c_idx in range(3)
                    ]
                }


        if type(self.annotations)==str:
            assert(os.path.exists(self.annotations))

        self.processed_annotations, self.annotations_metadata = self.load_annotations(self.annotations)

    def extract_meta_dict(self, annotations):

        if not type(annotations)==list:
            annotations = [annotations]

        ann_metadata = []
        for a in annotations:
            if hasattr(a,'to_dict'):
                a = a.to_dict()
                
            if 'properties' in a:
                ann_metadata.append({
                    'name': a['properties']['name'],
                    '_id': a['properties']['_id'] 
                })
            elif 'image_path' in a:
                ann_metadata.append(a)

        return ann_metadata

    def load_annotations(self, annotations):

        geojson_annotations = []
        annotations_metadata = []
        if not annotations is None:
            if type(annotations)==str:
                new_loaded_annotations = load_annotations(annotations)
                if not new_loaded_annotations is None:
                    if not type(new_loaded_annotations)==list:
                        new_loaded_annotations = [new_loaded_annotations]

                    geojson_annotations.extend(new_loaded_annotations)
                    annotations_metadata.extend(self.extract_meta_dict(new_loaded_annotations))
                else:
                    print(f'Unrecognized annotation format: {annotations}')

            elif hasattr(annotations,"to_dict"):
                geojson_annotations.append(annotations.to_dict())
                annotations_metadata.extend(self.extract_meta_dict(annotations))

            elif type(annotations)==list:
                for n in annotations:
                    processed_anns = []
                    if not n is None:
                        if hasattr(n,"to_dict"):
                            processed_anns.append(n.to_dict())
                        elif type(n)==dict:
                            if 'annotation' in n:
                                converted = histomics_to_geojson(n)
                                processed_anns.append(converted)
                            else:
                                processed_anns.append(n)
                        elif type(n)==str:
                            loaded_anns = load_annotations(n)
                            if not loaded_anns is None:
                                if type(loaded_anns)==list:
                                    processed_anns.extend(loaded_anns)
                                elif type(loaded_anns)==dict:
                                    processed_anns.append(loaded_anns)
                            else:
                                print(f'Unrecognized Format: {n}')

                        elif type(n)==np.ndarray:
                            print(f'Found annotations of type: {type(n)}, make sure to specify if this is an overlay image (use fusion_tools.SlideImageOverlay) or a label mask (use fusion_tools.utils.shapes.load_label_mask)')
                        else:
                            print(f'Unknown annotations type found: {n}')
                    
                    geojson_annotations.extend(processed_anns)
                    annotations_metadata.extend(self.extract_meta_dict(processed_anns))
                        
            elif type(annotations)==dict:
                if 'annotation' in annotations:
                    converted_annotations = histomics_to_geojson(annotations)
                    geojson_annotations.append(converted_annotations)
                    annotations_metadata.extend(self.extract_meta_dict([converted_annotations]))
                else:
                    geojson_annotations.append(annotations)
                    annotations_metadata.extend(self.extract_meta_dict([annotations]))

        return geojson_annotations, annotations_metadata    


class TileServer:
    """Components which pull information from a slide(s)
    """
    
    def get_slide_urls(slide_dict,**kwargs):
        raise NotImplementedError


class LocalTileServer(TileServer):
    """Tile server from image saved locally. Uses large-image to read and parse image formats (default: [common])
    """
    def __init__(self,
                 database: Union[fusionDB,None] = None,
                 tile_server_port:int = 8050,
                 host: str = 'localhost',
                 protocol: str = 'http',
                 jupyter_server_url: Union[str,None] = None,
                 cors_options: dict = {}
                 ):
        """Constructor method
        """

        self.database = database
        self.tile_server_port = tile_server_port
        self.host = host
        self.protocol = protocol
        self.cors_options = cors_options
        self.jupyter_server_url = jupyter_server_url

        if self.jupyter_server_url is None or self.jupyter_server_url=='':
            self.access_url = f'{self.protocol}://{self.host}:{self.tile_server_port}'
        else:
            self.access_url = self.jupyter_server_url

        self.app = FastAPI(
            title = 'TileServer',
            description = 'Locally deployed tileserver for high-resolution microscopy images',
            version = '3.6.42'
        )

        self.app.add_middleware(
            CORSMiddleware,
            allow_origins=self.cors_options.get('allow_origins',['*']),
            allow_methods=self.cors_options.get('allow_methods',['GET','OPTIONS']),
            allow_headers = self.cors_options.get('allow_headers',['*']),
            expose_headers = self.cors_options.get('expose_headers',['*']),
            allow_credentials = self.cors_options.get('allow_credentials',False)
        )

        self.router = APIRouter()
        self.router.add_api_route('/',self.root,methods=["GET","OPTIONS"])
        self.router.add_api_route('/ids',self.get_ids,methods=["GET","OPTIONS"])
        self.router.add_api_route('/names',self.get_names,methods=["GET","OPTIONS"])
        self.router.add_api_route('/{id}/info',self.get_id_info,methods=["GET","OPTIONS"])
        self.router.add_api_route('/{id}/tiles/{z}/{x}/{y}',self.get_tile,methods=["GET","OPTIONS"])
        self.router.add_api_route('/{id}/image_metadata',self.get_image_metadata,methods=["GET","OPTIONS"])
        self.router.add_api_route('/{id}/metadata',self.get_metadata,methods=["GET","OPTIONS"])
        self.router.add_api_route('/{id}/tiles/region',self.get_region,methods=["GET","OPTIONS"])
        self.router.add_api_route('/{id}/tiles/thumbnail',self.get_thumbnail,methods=["GET","OPTIONS"])
        self.router.add_api_route('/{id}/annotations',self.get_annotations,methods=["GET","OPTIONS"])
        self.router.add_api_route('/{id}/annotations/metadata',self.get_annotations_metadata,methods=["GET", "OPTIONS"])
        self.router.add_api_route('/{id}/annotations/data/list',self.get_annotations_property_keys,methods=["GET","OPTIONS"])
        self.router.add_api_route('/{id}/annotations/data',self.get_annotations_property_data,methods=["GET","OPTIONS"])

    def __str__(self):
        return f'TileServer class to {self.host}:{self.tile_server_port}'
    
    def add_new_image(self,
        new_image_id:str, 
        new_image_path:str, 
        new_annotations:Union[str,list,dict,None] = None, 
        new_metadata:Union[dict,None] = None, 
        new_image_style:Union[dict,None] = None,
        new_image_public: bool = False,
        session_id: Union[str,None] = None,
        user_id: Union[str,None] = None
        ):

        # Verifying filepaths and loading annotations
        new_local_item = Slide(
            image_filepath=new_image_path,
            annotations = new_annotations,
            metadata = new_metadata,
            image_style=new_image_style,
            public = new_image_public
        )

        self.add_new_slide(new_local_item,session_id,user_id)

    def add_new_slide(self, slide_id: str, slide_obj: Slide, session_id: Union[str,None] = None, user_id: Union[str,None] = None):

        slide_name = slide_obj.image_filepath.split(os.sep)[-1]
        # Adding information to database
        self.database.add_slide(
            slide_id = slide_id,
            slide_name = slide_name,
            metadata = slide_obj.metadata,
            image_metadata = slide_obj.image_metadata,
            image_filepath = slide_obj.image_filepath,
            annotations_metadata = slide_obj.annotations_metadata,
            annotations = slide_obj.processed_annotations,
            vis_session_id = session_id,
            user_id = user_id,
            public = slide_obj.public
        )

        # If this is not a public slide, add to UserAccess table
        #if not slide_obj.public:

        #    self.database.add_access(
        #        item_id = slide_id,
        #        user_id = user_id
        #    )

    def root(self):
        return {'message': "Oh yeah, now we're cooking"}

    async def get_item(self,item_id:str, token:Union[str,None] = None):
        """Grabbing item from database

        :param item_id: String uuid for local item
        :type item_id: str
        :return: Item instance
        :rtype: None
        """

        if not token is None:
            user_filter = {
                'user': {
                    'token': token
                }
            }
        else:
            user_filter = {}

        image_item = await self.database.search(
            search_kwargs = {
                'type': 'item',
                'filters': {
                    'item': {
                        'id': item_id
                    } 
                }| user_filter
            }
        )

        return image_item
    
    async def get_tile_source(self,item_id:str,style:Union[str,None]=None, token: Union[str,None] = None):
        """Getting large-image tile source for a given id+style combo. 

        :param item_id: String uuid for local image
        :type item_id: str
        :param style: Style dict for large-image reader, specifying bands/palettes, etc., defaults to None
        :type style: Union[str,None], optional
        :return: Tile source
        :rtype: None
        """

        image_item = await asyncio.gather(self.get_item(item_id,token))
        if len(image_item[0])==0:
            return None
        else:
            image_item = image_item[0][0]
            image_filepath = image_item.get('filepath')

            tile_source = large_image.open(
                image_filepath,
                style = json.loads(style) if type(style)==str else None
            )

            return tile_source
        
    async def get_item_annotations(self, item_id:str, request: Request = None):
        """Loading annotations from item database

        :param item_id: String uuid for local image
        :type item_id: str
        """
        user_filter = {}
        if not request is None:
            if request.query_params.get('token'):
                user_filter = {
                    'user': {
                        'token': request.query_params.get('token')
                    }
                }

        item_annotations = []

        item_layers = await self.database.search(
            search_kwargs = {
                'type': 'layer',
                'filters': {
                    'item': {
                        'id': item_id
                    }
                } | user_filter
            }
        )

        for l in item_layers:
            layer_name = l.get('name')
            layer_id = l.get('id')

            #TODO: Check if this also needs the user token
            layer_structures = await self.database.search(
                search_kwargs={
                    'type': 'structure',
                    'filters': {
                        'layer': {
                            'id': layer_id
                        } 
                    }| user_filter
                }
            )

            if len(layer_structures)>0:
                item_annotations.append(
                    {
                        'type': 'FeatureCollection',
                        'properties': {
                            'name': layer_name,
                            '_id': layer_id
                        },
                        'features': [
                            {
                                'type': 'Feature',
                                'geometry': s.get('geom'),
                                'properties': s.get('properties')
                            }
                            for s in layer_structures
                        ]
                    }
                )
            else:
                #TODO: Check if this needs a user_token argument
                # This could be an ImageOverlay layer
                image_overlays = await self.database.search(
                    search_kwargs = {
                        'type': 'image_overlay',
                        'filters': {
                            'layer': {
                                'id': layer_id
                            }
                        } | user_filter
                    }
                )

                if len(image_overlays)>0:
                    item_annotations.extend(
                        [
                            i.to_dict() for i in image_overlays
                        ]
                    )

        return item_annotations

    def get_names(self, request: Request):
        """Get names of items in fusionDB

        :return: Message dictionary containing list of all item names
        :rtype: dict
        """

        item_names = self.database.get_names(
            table_name = 'item',
        )

        return {'message': item_names}

    def get_ids(self):
        """Get all available ids in the database
        """
        item_ids = self.database.get_ids(
            table_name = 'item'
        )

        return {'message': item_ids}
    
    async def get_id_info(self, id:str):
        """Get info for a single id
        """
        image_item = await asyncio.gather(self.get_item(id))

        if len(image_item[0])>0:
            item_info = image_item[0][0]
            return Response(
                content = json.dumps(item_info,default=str),
                media_type = 'application/json'
            )
        else:
            return Response(
                content = 'invalid image id',
                media_type='application/json',
                status_code=400,
            )
    
    def get_item_names_ids(self, filters = None, size = None, offset = 0):
        """Get list of names and ids of all locally stored images in this tileserver
        """

        item_names_ids = self.database.search(
            search_kwargs={
                'type': 'item',
                'filters': filters
            },
            size = size,
            offset = offset
        )

        item_names_ids = [
            {
                'name': i.name,
                'id': i.id
            }
            for i in item_names_ids
        ]

        return item_names_ids

    @staticmethod
    def get_slide_urls(slide_dict,user_token,standalone = False):
        
        slide_id = slide_dict.get('id')
        url = slide_dict.get('url')

        if not standalone:
            slide_url_dict = {
                'tiles': f'{url}/tileserver/{slide_id}/tiles/'+'{z}/{x}/{y}',
                'regions': f'{url}/tileserver/{slide_id}/tiles/region',
                'image_metadata': f'{url}/tileserver/{slide_id}/image_metadata',
                'metadata': f'{url}/tileserver/{slide_id}/metadata',
                'annotations': f'{url}/tileserver/{slide_id}/annotations',
                'annotations_metadata':f'{url}/tileserver/{slide_id}/annotations/metadata',
                'annotations_region':f'{url}/tileserver/{slide_id}/annotations'
            }
        else:
            slide_url_dict = {
                'tiles': f'{url}/{slide_id}/tiles/'+'{z}/{x}/{y}',
                'regions': f'{url}/{slide_id}/tiles/region',
                'image_metadata': f'{url}/{slide_id}/image_metadata',
                'metadata': f'{url}/{slide_id}/metadata',
                'annotations': f'{url}/{slide_id}/annotations',
                'annotations_metadata':f'{url}/{slide_id}/annotations/metadata',
                'annotations_region':f'{url}/{slide_id}/annotations'
            }

        if not user_token is None:
            for k,v in slide_url_dict.items():
                slide_url_dict[k] = v + f'?token={user_token}'

        return slide_url_dict

    def get_tiles_url(self,slide_id):
        tiles_url = f'{self.protocol}://{self.host}:{self.tile_server_port}/{slide_id}/tiles/'+'{z}/{x}/{y}'

        return tiles_url

    def get_regions_url(self,slide_id):
        regions_url = f'{self.protocol}://{self.host}:{self.tile_server_port}/{slide_id}/tiles/region'

        return regions_url

    def get_annotations_url(self,slide_id):
        annotations_url = f'{self.protocol}://{self.host}:{self.tile_server_port}/{slide_id}/annotations'

        return annotations_url

    def get_annotations_metadata_url(self,slide_id):
        annotations_metadata_url = f'{self.protocol}://{self.host}:{self.tile_server_port}/{slide_id}/annotations/metadata'

        return annotations_metadata_url

    def get_metadata_url(self,slide_id):
        metadata_url = f'{self.protocol}://{self.host}:{self.tile_server_port}/{slide_id}/metadata'

        return metadata_url
        
    def get_image_metadata_url(self,slide_id):
        image_metadata_url = f'{self.protocol}://{self.host}:{self.tile_server_port}/{slide_id}/image_metadata'

        return image_metadata_url

    async def get_tile(self,id:str,z:int, x:int, y:int, style:Union[None,str] = None, request:Request = None):
        """Tiles endpoint, returns an image tyle based on provided coordinates

        :param id: Local item id
        :type id: str
        :param z: Zoom level for tile
        :type z: int
        :param x: X tile coordinate
        :type x: int
        :param y: Y tile coordinate
        :type y: int
        :param style: Additional style arguments to pass to large-image, defaults to {}
        :type style: dict, optional
        :return: Image tile containing bytes encoded pixel information
        :rtype: Response
        """

        # large-image getTile doesn't do anything with negative zoom or tile coordinates
        if any([i<0 for i in [z,x,y]]):
            return Response(status_code=200)

        token = None
        if not request is None:
            if request.query_params.get('token'):
                token = request.query_params.get('token')

        tile_source = await asyncio.gather(self.get_tile_source(id,style, token))
        tile_source = tile_source[0]
        if tile_source is None:
            return Response(
                content = 'invalid image id', 
                media_type='application/json',
                status_code=400,
            )

        tile_metadata = tile_source.getMetadata()

        try:
            raw_tile = tile_source.getTile(
                x = x,
                y = y,
                z = z,
            )
    
        except large_image.exceptions.TileSourceXYZRangeError:
            # This error appears for any negative tile coordinates
            raw_tile = np.zeros(
                (
                    tile_metadata['tileHeight'],
                    tile_metadata['tileWidth']
                ),
                dtype=np.uint8
            ).tobytes()

        return Response(
            content = raw_tile, 
            media_type='image/png',
        )
    
    async def get_image_metadata(self,id:str, request:Request = None):
        """Getting large-image metadata for image

        :return: Dictionary containing metadata for local image
        :rtype: Response
        """

        token = None
        if not request is None:
            if request.query_params.get('token'):
                token = request.query_params.get('token')


        image_item = await asyncio.gather(self.get_item(id,token))
        if len(image_item[0])>0:
            image_meta = image_item[0][0].get('image_meta',{})
            if not image_meta is None:
                return Response(
                    content = json.dumps(image_meta),
                    media_type = 'application/json',
                )
            else:
                return Response(
                    content = json.dumps({}),
                    media_type = 'application/json',
                )
        else:
            return Response(
                content = 'invalid image id',
                media_type='application/json', 
                status_code=400,
            )
        
    async def get_metadata(self, id:str, request: Request = None):
        """Getting metadata associated with slide/case/patient

        :param image: Index of local image
        :type image: int
        """

        token = None
        if not request is None:
            if request.query_params.get('token'):
                token = request.query_params.get('token')

        image_item = await asyncio.gather(self.get_item(id, token))

        if len(image_item[0])>0:
            item_meta = image_item[0][0].get('meta',{})
            if not item_meta is None:
                return Response(
                    content = json.dumps(item_meta),
                    media_type = 'application/json',
                )
            else:
                return Response(
                    content = json.dumps({}),
                    media_type = 'application/json',
                )
        else:
            return Response(
                content = 'invalid image id',
                media_type='application/json',
                status_code=400,
            )
    
    async def get_region(self, id:str, top: int, left: int, bottom:int, right:int,style:Union[None,str] = None):
        """
        Grabbing a specific region in the image based on bounding box coordinates
        """
        """Grabbing a specific region of the image based on bounding box coordinates

        :return: Image region (bytes encoded)
        :rtype: Response
        """
        token = None
        if not request is None:
            if request.query_params.get('token'):
                token = request.query_params.get('token')

        tile_source = await asyncio.gather(self.get_tile_source(id,style, token))
        tile_source = tile_source[0]
        
        if tile_source is None:
            return Response(
                content = 'invalid image id',
                media_type = 'application/json',
                status_code = 400,
            )

        image_region, mime_type = tile_source.getRegion(
            region = {
                'left': left,
                'top': top,
                'right': right,
                'bottom': bottom
            },
        )

        return Response(
            content = image_region,
            media_type = 'image/png',
        )

    async def get_thumbnail(self, id:str, style:Union[None,str] = None, request: Request = None):
        """Grabbing an image thumbnail

        :param id: Unique id for locally stored item
        :type id: str
        """
        token = None
        if not request is None:
            if request.query_params.get('token'):
                token = request.query_params.get('token')

        tile_source = await asyncio.gather(self.get_tile_source(id,style, token))
        tile_source = tile_source[0]

        if tile_source is None:
            return Response(
                content = 'invalid image id', 
                media_type = 'application/json', 
                status_code=400,
            )

        thumbnail, mime_type = tile_source.getThumbnail(encoding='PNG')
        
        return Response(
            content = thumbnail, 
            media_type = 'image/png',
        )

    async def get_annotations(self,id:str, top:Union[int,None]=None, left:Union[int,None]=None, bottom: Union[int,None]=None, right: Union[int,None]=None, request: Request = None):
        """Getting annotations for a given item id, optionally specifying a region within which to grab annotations.

        :param id: String uuid for local image
        :type id: str
        :param top: Y-coordinates for top-left corner of region, defaults to None
        :type top: Union[int,None], optional
        :param left: X-coordinates for top-left corner of region, defaults to None
        :type left: Union[int,None], optional
        :param bottom: Y-coordinates for bottom-right corner of region, defaults to None
        :type bottom: Union[int,None], optional
        :param right: X-coordinates for bottom-right corner of region, defaults to None
        :type right: Union[int,None], optional
        :return: Annotations for item (optionally within a specified region)
        """

        image_annotations = await asyncio.gather(self.get_item_annotations(id, request))
        
        if len(image_annotations)>0:
            if all([i is None for i in [top,left,bottom,right]]):
                # Returning all annotations by default
                return Response(
                    content = json.dumps(image_annotations),
                    media_type='application/json',
                )
            else:
                # Parsing region of annotations:
                if all([not i is None for i in [top,left,bottom,right]]):
                    image_region_anns = []
                    # Shapely box requires minx, miny, maxx, maxy
                    query_poly = box(left,top,right,bottom)

                    if type(image_annotations)==dict:
                        image_anns = [image_annotations]
                    for ann in image_anns:
                        if detect_image_overlay(ann):
                            image_bounds_box = box(*ann['image_bounds'])
                            if image_bounds_box.intersects(query_poly):
                                image_region_anns.append(ann)
                                # If it doesn't intersect should it return anything?

                        elif detect_geojson(ann):
                            
                            if type(ann)==dict:
                                filtered_anns = structures_within_poly(
                                    original = ann,
                                    query= query_poly
                                )
                                if len(filtered_anns['features'])>0:
                                    filtered_anns['properties'] = ann['properties']
                                    image_region_anns.append(filtered_anns)
                                else:
                                    image_region_anns.append({
                                        'type': 'FeatureCollection',
                                        'features':[],
                                        'properties':ann['properties']
                                    })
                            elif type(ann)==list:
                                for g in ann:
                                    filtered_g = structures_within_poly(
                                        original = g,
                                        query=query_poly
                                    )
                                    if len(filtered_g['features'])>0:
                                        filtered_g['properties'] = g['properties']
                                        image_region_anns.append(filtered_g)
                                    else:
                                        image_region_anns.append({
                                            'type': 'FeatureCollection',
                                            'features':[],
                                            'properties':g['properties']
                                        })

                        else:
                            print(f'Unrecognized annotation format found for image: {id}')
                    return Response(
                        content = json.dumps(image_region_anns), 
                        media_type='application/json',
                    )

        else:
            return Response(
                content = 'invalid image id',
                media_type = 'application/json', 
                status_code = 400,
            )

    async def get_annotations_metadata(self,id:str, request: Request):
        """Getting metadata for annotations for an item

        :param id: String uuid for locally stored image
        :type id: str
        :return: Metadata associated with annotations for that image
        """

        token = None
        if not request is None:
            if request.query_params.get('token'):
                token = request.query_params.get('token')


        image_item = await asyncio.gather(self.get_item(id, token))
        if len(image_item[0])==0:
            return Response(
                content = 'invalid image id',
                media_type = 'application/json',
                status_code = 400,
            )
        
        image_item = image_item[0][0]

        ann_meta = image_item.get('ann_meta',[])

        return Response(
            content = json.dumps(ann_meta),
            media_type = 'application/json'
        )

    async def get_annotations_property_keys(self,id:str, request: Request):
        """Getting the names of properties stored in an image's annotations

        :param id: String uuid for locally stored image
        :type id: int
        """
        #print(f'id from get_annotations_property_keys: {id}')
        image_anns = await self.get_item_annotations(id, request)
        property_list = []
        property_names = []
        for a in image_anns:
            features = a.get('features',[])
            a_id = a.get('properties',{}).get('_id',None)
            a_name = a.get('properties',{}).get('name',None)

            if 'annotation.name' in property_names:
                v_info = property_list[property_names.index('annotation.name')]
                if not v_val in v_info['distinct']:
                    v_info['distinct'].append(v_val)
                    v_info['distinctcount'] += 1
                
                v_info['count'] +=1
                property_list[property_names.index('annotation.name')] = v_info

            else:
                property_names.append('annotation.name')
                
                v_info = {
                    'key': 'annotation.name',
                    'title': 'Annotation Name',
                    'count': 1,
                    'distinct': [a_name],
                    'distinctcount': 1
                }
                property_list.append(v_info)

            if 'annotation.id' in property_names:
                v_info = property_list[property_names.index('annotation.id')]
                if not v_val in v_info['distinct']:
                    v_info['distinct'].append(v_val)
                    v_info['distinctcount'] += 1

                v_info['count'] += 1

                property_list[property_names.index('annotation.id')] = v_info

            else:
                property_names.append('annotation.id')
                
                v_info = {
                    'key': 'annotation.id',
                    'title': 'Annotation ID',
                    'count': 1,
                    'distinct': [a_id],
                    'distinctcount': 1
                }
                property_list.append(v_info)


            for f in features:
                f_props = f.get('properties')
                if not f_props is None:
                    f_main_keys = list(f_props.keys())
                    for f_m in f_main_keys:
                        if type(f_props[f_m]) in [list,dict]:
                            values_list = extract_nested_prop(f_props[f_m],4)
                        elif type(f_props[f_m]) in [str,int,float]:
                            values_list = [{f_m: f_props[f_m]}]
                        
                        for v in values_list:
                            v_key = list(v.keys())[0]
                            v_val = list(v.values())[0]

                            if v_key in property_names:
                                v_info = property_list[property_names.index(v_key)]
                                if type(v_val)==str:
                                    if not v_val in v_info['distinct']:
                                        v_info['distinct'].append(v_val)
                                        v_info['distinctcount'] += 1
                                    
                                elif type(v_val) in [int,float]:
                                    if v_val>v_info['max']:
                                        v_info['max'] = v_val
                                    elif v_val<v_info['min']:
                                        v_info['min'] = v_val
                                    
                                v_info['count'] += 1

                                property_list[property_names.index(v_key)] = v_info
                            else:
                                property_names.append(v_key)
                                
                                v_info = {
                                    'key': v_key.lower().replace(' --> ','.'),
                                    'title': v_key,
                                    'count': 1
                                }

                                if type(v_val)==str:
                                    v_info['type'] = 'string'
                                    v_info['distinct'] = [v_val]
                                    v_info['distinctcount'] = 1
                                else:
                                    v_info['type'] = 'number'
                                    v_info['max'] = v_val
                                    v_info['min'] = v_val

                                property_list.append(v_info)
                                    

        return Response(
            content = json.dumps(property_list),
            media_type='application/json',
            status_code=200,
        )

    async def get_annotations_property_data(self,id:str,include_keys:Union[str,None] = None,include_anns:Union[str,None] = None, request: Request = None):
        """Getting data from annotations of specified image, attempting to mirror output of https://github.com/girder/large_image/blob/master/girder_annotation/girder_large_image_annotation/utils/__init__.py

        :param id: String uuid for locally stored image.
        :type id: int
        :param include_keys: List of property names to grab from annotations
        :type include_keys: list
        :param include_anns: Which annotations to include (name/id or __all__ or list)
        :type include_anns: Union[str,list,None]
        """

        token = None
        if not request is None:
            if request.query_params.get('token'):
                token = request.query_params.get('token')

                
        image_item = await self.get_item(id, token)[0]
        image_anns = await self.get_item_annotations(id,request)

        include_keys = include_keys.split(',')

        if include_keys is None:
            include_keys = []

        if include_anns is None:
            include_anns = '__all__'

        bbox_list = ['bbox.x0','bbox.y0','bbox.x1','bbox.y1']

        data_list = []
        for a in image_anns:
            features = a.get('features',[])
            a_id = a.get('properties',{}).get('_id',None)
            a_name = a.get('properties',{}).get('name',None)

            # Accepting either annotation layer name or id
            if not include_anns=='__all__':
                if type(include_anns)==str:
                    if not include_anns==a_id and not include_anns==a_name:
                        continue
                elif type(include_anns)==list:
                    if not a_id in include_anns and not a_name in include_anns:
                        continue
                        
            for f in features:
                f_props = f.get('properties')
                f_bbox = list(shape(f['geometry']).bounds)
                f_props_cols = []
                if not f_props is None:
                    for k in include_keys:
                        k = k.replace('data.','')
                        # Need to specify non-feature keys
                        if k in ['annotation.id','annotation.name','item.id','item.name']:
                            if k=='annotation.id':
                                f_props_cols.append(a_id)
                            elif k=='annotation.name':
                                f_props_cols.append(a_name)
                            elif k=='item.id':
                                f_props_cols.append(str(id))
                            elif k=='item.name':
                                f_props_cols.append(image_item.get('name'))

                        elif k in bbox_list:
                            # Adding bounding box coordinates
                            f_props_cols.append(f_bbox[bbox_list.index(k)])

                        else:
                            # Getting keys and nested keys
                            if '-->' in k:
                                f_sub_keys = k.split('-->')
                                f_sub_props = deepcopy(f_props)
                                for sk in f_sub_keys:
                                    if not f_sub_props is None:
                                        f_sub_props = f_sub_props.get(sk)
                            else:
                                f_sub_props = f_props.get(k)

                            # Converting to float if able
                            if f_sub_props is not None:
                                try:
                                    f_sub_props = float(f_sub_props)
                                    f_props_cols.append(f_sub_props)
                                except ValueError:
                                    f_props_cols.append(f_sub_props)
                            else:
                                f_props_cols.append(f_sub_props)
                else:
                    f_props_cols = [None]*len(include_keys)

                data_list.append(f_props_cols)

        return Response(
            content = json.dumps({'data': data_list, 'columns': include_keys}),
            media_type='application/json',
            status_code=200,
        )

    def start(self):
        """Starting tile server instance on a provided port

        :param port: Tile server port from which tiles are accessed, defaults to '8050'
        :type port: str, optional
        """
        self.app.include_router(self.router)
        
        # Enabling CORS (https://fastapi.tiangolo.com/tutorial/cors/#use-corsmiddleware)
        self.app.add_middleware(
            CORSMiddleware,
            allow_origins=self.cors_options.get('allow_origins',['*']),
            allow_methods=self.cors_options.get('allow_methods',['GET','OPTIONS']),
            allow_headers = self.cors_options.get('allow_headers',['*']),
            expose_headers = self.cors_options.get('expose_headers',['*']),
            allow_credentials = self.cors_options.get('allow_credentials',False)
        )

        uvicorn.run(self.app,host=self.host,port=self.tile_server_port)


class DSATileServer(TileServer):
    """Use for linking visualization with remote tiles API (DSA server)
    """
    def __init__(self,
                 api_url: str,
                 item_id: str,
                 user_token: Union[str,None]=None):
        """Constructor method

        :param api_url: URL for DSA API (ends in /api/v1)
        :type api_url: str
        :param item_id: Girder item Id to get tiles from
        :type item_id: str
        """

        self.base_url = api_url
        self.item_id = item_id

        #TODO: Add some method for appending the user_token to these URLs (Might be better to save this on the 
        # component side so that that property can be dynamic)
        if user_token is None:
            info_url = f'{api_url}/item/{item_id}'

        else:
            info_url = f'{api_url}/item/{item_id}?token={user_token}'

        item_info = requests.get(info_url).json()
        self.item_metadata = item_info['meta']
        self.name = item_info['name']

        self.metadata_url = info_url

        self.tiles_url = f'{api_url}/item/{item_id}/tiles/zxy/'+'{z}/{x}/{y}'
        self.regions_url = f'{api_url}/item/{item_id}/tiles/region'
        self.image_metadata_url = f'{api_url}/item/{item_id}/tiles'

        self.tiles_metadata = requests.get(
            f'{api_url}/item/{item_id}/tiles'
        ).json()

        #TODO: Add some method here for accessing the /annotation/{ann_id}/region endpoint?
        self.annotations_url = f'{api_url}/annotation/item/{item_id}'

        if user_token is None:
            self.annotations_metadata_url = f'{api_url}/annotation?itemId={item_id}'
        else:
            self.annotations_metadata_url = f'{api_url}/annotation?token={user_token}&itemId={item_id}'

        annotations_metadata = requests.get(self.annotations_metadata_url).json()
        self.annotations_geojson_url = [f'{api_url}/annotation/{a["_id"]}/geojson' for a in annotations_metadata]

        self.annotations_region_url = f'{api_url}/annotation/'

    def __str__(self):
        return f'DSATileServer for {self.base_url}'

    @staticmethod
    def get_slide_urls(slide_dict,user_token):
        
        item_id = slide_dict.get('remote_id')
        item_url = slide_dict.get('url')

        slide_url_dict = {
            'tiles': f'{item_url}/item/{item_id}/tiles/zxy/'+'{z}/{x}/{y}',
            'regions': f'{item_url}/item/{item_id}/tiles/region',
            'image_metadata': f'{item_url}/item/{item_id}/tiles',
            'metadata': f'{item_url}/item/{item_id}',
            'annotations': f'{item_url}/annotation/item/{item_id}',
            'annotations_metadata':f'{item_url}/annotation?itemId={item_id}',
            'annotations_region':f'{item_url}/annotation'
        }

        if not user_token is None:
            for k,v in slide_url_dict.items():
                if not '?' in v:
                    slide_url_dict[k] = v + f'?token={user_token}'
                else:
                    slide_url_dict[k] = v + f'&token={user_token}'

        # Adding annotations_geojson_url?
        annotations_metadata = requests.get(slide_url_dict.get('annotations_metadata')).json()

        annotations_geojson_urls = []
        for a in annotations_metadata:
            req_str = f'{item_url}/annotation/{a.get("_id")}/geojson'
            if not user_token is None:
                req_str += f'?token={user_token}'

            annotations_geojson_urls.append(
                req_str
            )

        slide_url_dict['annotations_geojson_url'] = annotations_geojson_urls

        return slide_url_dict
    
    @staticmethod
    def get_tile_iterator(slide_url_dict, tile_size):
        
        slide_image_metadata = requests.get(slide_url_dict.get('image_metadata')).json()
        slide_height, slide_width = slide_image_metadata.get('sizeY'), slide_image_metadata.get('sizeX')
        # minx, miny, maxx, maxy
        available_bbox = [
            0,0,
            slide_width,slide_height
        ]

        x_start = np.maximum(int(tile_size.get('width')/2),int(available_bbox[0]+(tile_size.get('width')/2)))
        y_start = np.maximum(int(tile_size.get('height')/2),int(available_bbox[1]+(tile_size.get('height')/2)))
        n_x = floor((available_bbox[2]-available_bbox[0])/tile_size.get('width'))
        n_y = floor((available_bbox[3]-available_bbox[1])/tile_size.get('height'))

        bbox_list = []
        for x in range(0,n_x):
            for y in range(0,n_y):

                # Adding column of bboxes
                bbox = [
                    int(x_start-(tile_size.get('width')/2)),
                    int(y_start-(tile_size.get('height')/2)),
                    int(x_start+(tile_size.get('width')/2)),
                    int(y_start+(tile_size.get('height')/2))
                ]

                bbox_list.append(bbox)

                y_start+=tile_size.get('height')

            bottom_row_bbox = [
                int(x_start - (tile_size.get('width')/2)),
                int(available_bbox[3]-tile_size.get('height')),
                int(x_start + (tile_size.get('width')/2)),
                int(available_bbox[3])
            ]

            bbox_list.append(bottom_row_bbox)
            x_start += tile_size.get('width')
            y_start = np.maximum(tile_size.get('height')/2,available_bbox[1]-(tile_size.get('height')/2))

        # Adding right-side column
        for y in range(0,n_y):
            right_column_bbox = [
                int(available_bbox[2] - tile_size.get('width')),
                int(y_start - (tile_size.get('height')/2)),
                int(available_bbox[2]),
                int(y_start + (tile_size.get('height')/2))
            ]

            bbox_list.append(right_column_bbox)
            y_start+=tile_size.get('height')

        # Adding bottom-right corner
        bottom_right_bbox = [
            int(available_bbox[2]-tile_size.get('width')),
            int(available_bbox[3]-tile_size.get('height')),
            int(available_bbox[2]),
            int(available_bbox[3])
        ]

        bbox_list.append(bottom_right_bbox)

        if '?' in slide_url_dict.get('regions'):
            sep_str = '&'
        else:
            sep_str = '?'

        bbox_region_urls = [
            f'{slide_url_dict.get("regions")}/{sep_str}left={b[0]}&top={b[1]}&right={b[2]}&bottom={b[3]}'
            for b in bbox_list
        ]

        return bbox_region_urls


class CustomTileServer(TileServer):
    """CustomTileServer component if using some other tiles endpoint (must pass tileSize and levels in dictionary)
    """
    def __init__(self,
                 tiles_url:str,
                 regions_url:str,
                 image_metadata: dict,
                 annotations_url: Union[str,None] = None,
                 name: str = None
                 ):
        """Constructor method

        :param tiles_url: URL to grab tiles from (ends in "/{z}/{x}/{y}")
        :type tiles_url: str
        :param regions_url: URL to grab image regions from
        :type regions_url: str
        :param image_metadata: Dictionary containing at least ['tileWidth','tileHeight','sizeX','sizeY','levels']
        :type image_metadata: dict
        """
        
        self.tiles_url = tiles_url
        self.regions_url = regions_url
        self.annotations_url = annotations_url
        self.name = name

        assert all([i in image_metadata for i in ['tileWidth','tileHeight','sizeX','sizeY','levels']])
        self.tiles_metadata = image_metadata

    def __str__(self):
        return f'CustomTileServer for {self.tiles_url}'

    def get_slide_urls(self, slide_dict):
        pass



"""

Components which extract data from the current SlideMap for formatted download

"""

import os
import sys
import json
import geojson
import geopandas as gpd
import numpy as np
import pandas as pd
import textwrap
import re
import uuid
import threading
import zipfile
from shutil import rmtree
from copy import deepcopy

from typing_extensions import Union
from shapely.geometry import box, shape
import plotly.express as px
import plotly.graph_objects as go
from umap import UMAP

from PIL import Image, ImageOps

from io import BytesIO
import requests

# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
import dash_leaflet as dl
import dash_leaflet.express as dlx
from dash import dcc, callback, ctx, ALL, MATCH, exceptions, Patch, no_update, dash_table
from dash.dash_table.Format import Format, Scheme
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
#import dash_treeview_antd as dta
from dash_extensions.enrich import DashBlueprint, html, Input, Output, State, PrefixIdTransform, MultiplexerTransform, BlockingCallbackTransform
from dash_extensions.javascript import Namespace, arrow_function

# fusion-tools imports
from fusion_tools.visualization.vis_utils import get_pattern_matching_value
from fusion_tools.utils.shapes import (
    find_intersecting, 
    extract_geojson_properties, 
    process_filters_queries,
    detect_histomics,
    histomics_to_geojson,
    export_annotations
)
from fusion_tools.utils.images import get_feature_image, write_ome_tiff, format_intersecting_masks
from fusion_tools.utils.stats import get_label_statistics, run_wilcox_rank_sum
from fusion_tools.components.base import Tool, MultiTool

import time


class DataExtractor(Tool):
    """
    """

    title = 'Data Extractor'
    description = 'Download select properties from indicated structures in the current slide.'

    def __init__(self):

        super().__init__()

        self.exportable_session_data = {
            'Slide Metadata':{
                'description': 'This is any information about the current slide including labels, preparation details, and image tile information.'
            },
            'Visualization Session':{
                'description': 'This is a record of your current Visualization Session which can be uploaded in the Dataset Builder page to reload.'
            }
        }


        self.exportable_data = {
            'Properties':{
                'formats': ['CSV','XLSX','JSON'],
                'description': 'These are all per-structure properties and can include morphometrics, cell composition, channel intensity statistics, labels, etc.'
            },
            'Annotations':{
                'formats': ['Histomics (JSON)','GeoJSON','Aperio XML'],
                'description': 'These are the boundaries of selected structures. Different formats indicate the type of file generated and imported into another tool for visualization and analysis.'
            },
            'Images & Masks':{
                'formats': ['OME-TIFF'],
                'description': 'This will be a zip file containing combined images and masks for selected structures.'
            },
            'Images': {
                'formats': ['OME-TIFF','TIFF','PNG','JPG'],
                'description': 'This will be a zip file containing images of selected structures. Note: If this is a multi-frame image and OME-TIFF is not selected, images will be rendered in RGB according to the current colors in the map.'
            },
            'Masks': {
                'formats': ['OME-TIFF','TIFF','PNG','JPG'],
                'description': 'This will be a zip file containing masks of selected structures. Note: If a Manual ROI is selected, masks will include all intersecting structures as a separate label for each.'
            }
        }

    def get_scale_factors(self, image_metadata: dict):
        """Function used to initialize scaling factors applied to GeoJSON annotations to project annotations into the SlideMap CRS (coordinate reference system)

        :return: x and y (horizontal and vertical) scale factors applied to each coordinate in incoming annotations
        :rtype: float
        """

        base_dims = [
            image_metadata['sizeX']/(2**(image_metadata['levels']-1)),
            image_metadata['sizeY']/(2**(image_metadata['levels']-1))
        ]

        #x_scale = (base_dims[0]*(240/image_metadata['tileHeight'])) / image_metadata['sizeX']
        #y_scale = -((base_dims[1]*(240/image_metadata['tileHeight'])) / image_metadata['sizeY'])

        x_scale = base_dims[0] / image_metadata['sizeX']
        y_scale = -(base_dims[1]) / image_metadata['sizeY']


        return x_scale, y_scale

    def gen_layout(self,session_data:dict):

        layout = html.Div([
            dbc.Card(
                dbc.CardBody([
                    dbc.Row(
                        html.H3(self.title)
                    ),
                    html.Hr(),
                    dbc.Row(
                        self.description
                    ),
                    html.Hr(),
                    html.Div([
                        dcc.Store(
                            id = {'type': 'data-extractor-store','index': 0},
                            storage_type = 'memory',
                            data = json.dumps({'selected_data': []})
                        ),
                        dbc.Modal(
                            id = {'type': 'data-extractor-download-modal','index': 0},
                            children = [],
                            is_open = False,
                            size = 'xl'
                        ),
                        dcc.Interval(
                            id = {'type': 'data-extractor-download-interval','index': 0},
                            disabled = True,
                            interval = 3000,
                            n_intervals = 0,
                            max_intervals=-1
                        )
                    ]),
                    dbc.Row([
                        dbc.Row([
                            dbc.Col(
                                html.H5('Download Session Data:'),
                                md = 5
                            ),
                            dbc.Col(
                                dcc.Dropdown(
                                    options = list(self.exportable_session_data.keys()),
                                    value = [],
                                    multi = False,
                                    placeholder = 'Select an option',
                                    id = {'type': 'data-extractor-session-data-drop','index': 0}
                                ),
                                md = 7
                            )
                        ]),
                        dbc.Row(
                            html.Div(
                                id = {'type': 'data-extractor-session-data-description','index': 0},
                                children = []
                            )
                        ),
                        dbc.Row(
                            dbc.Button(
                                'Download Session Data',
                                id = {'type': 'data-extractor-download-session-data','index': 0},
                                n_clicks = 0,
                                className = 'd-grid col-12 mx-auto',
                                color = 'primary',
                                disabled = True
                            )
                        )
                    ],style = {'marginBottom':'10px'}),
                    html.Hr(),
                    dbc.Row([
                        dbc.Col(
                            html.H5('Select which structures to extract data from.'),
                            md = 9
                        ),
                        dbc.Col([
                            html.A(
                                html.I(
                                    className = 'fa-solid fa-rotate fa-xl',
                                    n_clicks = 0,
                                    id = {'type': 'data-extractor-refresh-icon','index': 0}
                                )
                            ),
                            dbc.Tooltip(
                                target = {'type': 'data-extractor-refresh-icon','index': 0},
                                children = 'Click to refresh available structures'
                            )
                        ],md = 3)
                    ],justify='left'),
                    dbc.Row([
                        dbc.Col([
                            dbc.Label('Current Structures:',html_for={'type':'data-extractor-current-structures-drop','index': 0})
                        ],md=3),
                        dbc.Col([
                            dcc.Dropdown(
                                options = [],
                                value = [],
                                multi = True,
                                placeholder = 'Structures in Slide',
                                id = {'type': 'data-extractor-current-structures-drop','index': 0}
                            )
                        ],md=9)
                    ]),
                    html.Hr(),
                    dbc.Row([
                        dbc.Col([
                            html.H5('Select what type of data you want to extract')
                        ])
                    ]),
                    dbc.Row([
                        dbc.Col(
                            dbc.Label('Available Data:',html_for={'type': 'data-extractor-available-data-drop','index': 0}),
                            md = 3
                        ),
                        dbc.Col(
                            dcc.Dropdown(
                                options = [],
                                value = [],
                                multi = True,
                                placeholder='Data',
                                id = {'type': 'data-extractor-available-data-drop','index':0}
                            ),
                            md = 9
                        )
                    ]),
                    html.Hr(),
                    html.Div(
                        id = {'type':'data-extractor-selected-data-parent','index': 0},
                        children = [
                            html.Div('Selected data descriptions will appear here')
                        ],
                        style = {'maxHeight': '100vh','overflow': 'scroll'}
                    ),
                    dbc.Row([
                        dcc.Loading([
                            dbc.Button(
                                'Download Selected Data',
                                className = 'd-grid col-12 mx-auto',
                                disabled = True,
                                color = 'primary',
                                id = {'type': 'data-extractor-download-button','index': 0}
                            ),
                            dcc.Download(
                                id = {'type': 'data-extractor-download','index': 0}
                            )
                        ])
                    ],style = {'marginTop':'10px'})
                ])
            )
        ])

        self.blueprint.layout = layout

    def get_callbacks(self):
        
        # Callback for updating current slide
        self.blueprint.callback(
            [
                Input({'type': 'map-annotations-info-store','index': ALL},'data')
            ],
            [
                Output({'type': 'data-extractor-current-structures-drop','index': ALL},'options'),
                Output({'type': 'data-extractor-available-data-drop','index': ALL},'options'),
                Output({'type': 'data-extractor-selected-data-parent','index': ALL},'children'),
                Output({'type': 'data-extractor-download-button','index': ALL},'disabled'),
                Output({'type':'data-extractor-store','index': ALL},'data')
            ],
            prevent_initial_call = True
        )(self.update_slide)

        # Callback for the refresh button being pressed
        self.blueprint.callback(
            [
                Input({'type': 'data-extractor-refresh-icon','index': ALL},'n_clicks')
            ],
            [
                Output({'type': 'data-extractor-current-structures-drop','index': ALL},'options'),
                Output({'type': 'data-extractor-available-data-drop','index': ALL},'options'),
                Output({'type': 'data-extractor-selected-data-parent','index': ALL},'children'),
                Output({'type': 'data-extractor-download-button','index': ALL},'disabled'),
                Output({'type': 'data-extractor-store','index': ALL},'data')
            ],
            [
                State({'type': 'feature-overlay','index': ALL},'name'),
                State({'type': 'map-marker-div','index': ALL},'children')
            ],
            prevent_initial_call = True
        )(self.refresh_data)

        # Callback for selecting a session data item
        self.blueprint.callback(
            [
                Input({'type':'data-extractor-session-data-drop','index': ALL},'value')
            ],
            [
                Output({'type': 'data-extractor-session-data-description','index': ALL},'children'),
                Output({'type': 'data-extractor-download-session-data','index': ALL},'disabled')
            ]
        )(self.update_session_data_description)

        # Callback for downloading session data item
        self.blueprint.callback(
            [
                Input({'type': 'data-extractor-download-session-data','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'data-extractor-session-data-drop','index': ALL},'value'),
                State({'type': 'map-slide-information','index':ALL},'data'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'data-extractor-download','index': ALL},'data')
            ]
        )(self.download_session_data)

        # Callback for updating selected data information and enabling download button
        self.blueprint.callback(
            [
                Input({'type': 'data-extractor-available-data-drop','index': ALL},'value')
            ],
            [
                State({'type': 'data-extractor-store','index': ALL},'data'),
                State({'type': 'map-slide-information','index': ALL},'data'),
                State({'type': 'channel-mixer-tab','index': ALL},'label')
            ],
            [
                Output({'type': 'data-extractor-selected-data-parent','index': ALL},'children'),
                Output({'type': 'data-extractor-download-button','index': ALL},'disabled'),
                Output({'type': 'data-extractor-store','index': ALL},'data')
            ],
            prevent_initial_call = True
        )(self.update_data_info)

        # Disabling channel selection if 'Use ChannelMixer' is checked
        self.blueprint.callback(
            [
                Input({'type': 'data-extractor-channel-mix-switch','index': MATCH},'checked')
            ],
            [
                State({'type': 'channel-mixer-tab','index':ALL},'label')
            ],
            [
                Output({'type': 'data-extractor-selected-data-channels','index': MATCH},'disabled'),
                Output({'type': 'data-extractor-selected-data-channels','index': MATCH},'value')
            ],
            prevent_initial_call = True
        )(self.disable_channel_selector)

        # Callback for downloading selected data and clearing selections
        self.blueprint.callback(
            [
                Input({'type': 'data-extractor-download-button','index': ALL},'n_clicks')
            ],
            [
                Output({'type': 'data-extractor-download-interval','index': ALL},'disabled'),
                Output({'type': 'data-extractor-current-structures-drop','index': ALL},'value'),
                Output({'type': 'data-extractor-available-data-drop','index': ALL},'value'),
                Output({'type': 'data-extractor-selected-data-parent','index': ALL},'children'),
                Output({'type': 'data-extractor-download-button','index': ALL},'disabled'),
                Output({'type': 'data-extractor-store','index': ALL},'data')
            ],
            [
                State({'type': 'data-extractor-current-structures-drop','index': ALL},'value'),
                State({'type': 'data-extractor-available-data-drop','index': ALL},'value'),
                State({'type': 'data-extractor-selected-data-format','index': ALL},'value'),
                State({'type': 'map-annotations-store','index':ALL},'data'),
                State({'type': 'map-marker-div','index': ALL},'children'),
                State({'type': 'map-slide-information','index': ALL},'data'),
                State({'type': 'channel-mixer-tab','index': ALL},'label'),
                State({'type': 'channel-mixer-tab','index': ALL},'label_style'),
                State({'type': 'data-extractor-channel-mix-switch','index': ALL},'checked'),
                State({'type': 'data-extractor-selected-data-channels','index': ALL},'value'),
                State({'type': 'data-extractor-selected-data-masks','index': ALL},'value')
            ],
            prevent_initial_call = True
        )(self.start_download_data)

        self.blueprint.callback(
            [
                Input({'type': 'data-extractor-download-interval','index': ALL},'n_intervals')
            ],
            [
                State({'type': 'data-extractor-store','index': ALL},'data')
            ],
            [
                Output({'type': 'data-extractor-store','index': ALL},'data'),
                Output({'type':'data-extractor-download-interval','index': ALL},'disabled'),
                Output({'type': 'data-extractor-download-modal','index': ALL},'is_open'),
                Output({'type': 'data-extractor-download-modal','index':ALL},'children'),
                Output({'type': 'data-extractor-download-interval','index':ALL},'n_intervals'),
                Output({'type': 'data-extractor-download','index': ALL},'data')
            ],
            prevent_initial_call = True
        )(self.update_download_data)
    
    def update_slide(self, new_annotations_info:list):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        new_annotations_info = json.loads(get_pattern_matching_value(new_annotations_info))
        new_structure_names = new_annotations_info['feature_names']

        available_structures_drop = [
            {'label': i, 'value': i, 'disabled': False}
            for i in new_structure_names
        ]
        if len(new_structure_names)>0:
            available_structures_drop += [{'label': 'Marked Structures','value': 'Marked Structures','disabled': True}]

        if len(available_structures_drop)>0:
            available_data_drop = [
                {'label': i, 'value': i, 'disabled': False}
                for i in self.exportable_data
            ]
        else:
            available_data_drop = []

        button_disabled = True
        new_data_extractor_store = json.dumps({'selected_data': []})

        return [available_structures_drop], [available_data_drop], [html.Div('Selected data descriptions will appear here')], [button_disabled], [new_data_extractor_store]

    def refresh_data(self, clicked, overlay_names, marker_div_children):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        marker_div_children = get_pattern_matching_value(marker_div_children)
        if not overlay_names is None:
            available_structures_drop = [
                {'label': i, 'value': i, 'disabled': False}
                for i in overlay_names
            ]
        else:
            available_structures_drop = []
        
        if not marker_div_children is None:
            available_structures_drop += [{'label': 'Marked Structures','value': 'Marked Structures','disabled': True if len(marker_div_children)==0 else False}]

        if not overlay_names is None or not marker_div_children is None:
            available_data_drop = [
                {'label': i, 'value': i, 'disabled': False}
                for i in self.exportable_data
            ]
        else:
            available_data_drop = []

        button_disabled = True

        new_data_extractor_store = json.dumps({'selected_data': []})

        return [available_structures_drop], [available_data_drop], [html.Div('Selected data descriptions will appear here')], [button_disabled],[new_data_extractor_store]

    def update_session_data_description(self, session_data_selection):
        
        if not any([i['value'] for i in ctx.triggered]):
            return ['Select a type of session data to see a description'], [True]
        
        session_data_selection = get_pattern_matching_value(session_data_selection)

        return [self.exportable_session_data[session_data_selection]['description']], [False]

    def download_session_data(self, clicked, session_data_selection, current_slide_info, session_data):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        session_data_selection = get_pattern_matching_value(session_data_selection)
        current_slide_info = json.loads(get_pattern_matching_value(current_slide_info))
        session_data = json.loads(session_data)

        if session_data_selection == 'Slide Metadata':
            if 'tiles_url' in current_slide_info:
                slide_tile_url = current_slide_info['tiles_url']
                current_slide_tile_urls = [i['tiles_url'] for i in session_data['current']]
                slide_idx = current_slide_tile_urls.index(slide_tile_url)

                slide_metadata = requests.get(session_data['current'][slide_idx]['metadata_url']).json()

                download_content = {'content': json.dumps(slide_metadata,indent=4),'filename': 'slide_metadata.json'}
            else:
                raise exceptions.PreventUpdate
        elif session_data_selection == 'Visualization Session':
            download_content = {'content': json.dumps(session_data,indent=4),'filename': 'fusion_visualization_session.json'}

        return [download_content]

    def update_data_info(self, selected_data, data_extract_store, slide_info_store, channel_mix_frames):

        if not any([i['value'] for i in ctx.triggered]):
            return [html.Div('Selected data descriptions will appear here')], [True], [no_update]
        
        selected_data = get_pattern_matching_value(selected_data)
        current_selected_data = json.loads(get_pattern_matching_value(data_extract_store))['selected_data']
        slide_info = json.loads(get_pattern_matching_value(slide_info_store))['tiles_metadata']

        channel_mix_opt = not channel_mix_frames is None and not channel_mix_frames==[]        

        def make_new_info(data_type, slide_info, channel_mix_opt):
            data_type_index = list(self.exportable_data.keys()).index(data_type)

            # Checking if this info card should also have a channel selector
            if 'Images' in data_type and any([i in slide_info for i in ['frames','channels','channelmap']]):
                show_channels = True

                if 'channels' in slide_info:
                    channel_names = slide_info['channels']
                else:
                    channel_names = [f'Channel {i+1}' for i in range(len(slide_info['frames']))]

            else:
                show_channels = False
                channel_names = ['red','green','blue']

            if "Masks" in data_type:
                show_mask_opts = True
                mask_opts = ['Structure Only','Intersecting']
            else:
                show_mask_opts = False
                mask_opts = []

            return_card = dbc.Card([
                dbc.CardHeader(data_type),
                dbc.CardBody([
                    dbc.Row([
                        dbc.Col(
                            dbc.Label('Download Format: '),
                            md = 3
                        ),
                        dbc.Col(
                            dcc.Dropdown(
                                options = self.exportable_data[data_type]['formats'],
                                value = self.exportable_data[data_type]['formats'][0],
                                multi = False,
                                placeholder = 'Select a format',
                                id = {'type': f'{self.component_prefix}-data-extractor-selected-data-format','index': data_type_index}
                            ),
                            md = 9
                        )
                    ]),
                    dbc.Row([
                        dbc.Col([
                            dmc.Switch(
                                label = 'Use ChannelMixer Colors',
                                checked = False,
                                description='Use colors selected in the ChannelMixer component, renders an RGB image.',
                                id = {'type': f'{self.component_prefix}-data-extractor-channel-mix-switch','index': data_type_index}
                            )
                        ])
                    ],style = {'display':'none'} if not channel_mix_opt else {}),
                    dbc.Row([
                        dbc.Col([
                            dbc.Label('Channels to save: ')
                        ], md = 3),
                        dbc.Col(
                            dcc.Dropdown(
                                options = channel_names,
                                value = channel_names,
                                multi = True,
                                id = {'type': f'{self.component_prefix}-data-extractor-selected-data-channels','index': data_type_index}
                            ),
                            md = 9
                        )
                    ],style = {'display': 'none'} if not show_channels else {}
                    ),
                    dbc.Row([
                        dbc.Col([
                            dbc.Label('Mask Options: ')
                        ],md=3),
                        dbc.Col(
                            dcc.Dropdown(
                                options = mask_opts,
                                value = mask_opts[0] if len(mask_opts)>0 else [],
                                multi = False,
                                id = {'type': f'{self.component_prefix}-data-extractor-selected-data-masks','index': data_type_index}
                            ),
                            md = 9
                        )
                    ],style = {'display': 'none'} if not show_mask_opts else {}
                    )
                ])
            ])

            return return_card

        # Which ones are in selected_data that aren't in current_selected_data
        if len(current_selected_data)>0 and len(selected_data)>0:
            new_selected = list(set(selected_data) - set(current_selected_data))
            info_return = Patch()
            if len(new_selected)>0:
                # Appending to the partial property update
                for n in new_selected:
                    info_return.append(make_new_info(n,slide_info,channel_mix_opt))

            else:
                # Removing some de-selected values
                removed_type = list(set(current_selected_data)-set(selected_data))
                rem_count = 0
                for r in removed_type:
                    del info_return[current_selected_data.index(r)-rem_count]
                    rem_count+=1

        elif len(current_selected_data)>0 and len(selected_data)==0:
            info_return = html.Div('Selected data descriptions will appear here')
        elif len(current_selected_data)==0 and len(selected_data)==1:
            info_return = Patch()
            info_return.append(make_new_info(selected_data[0],slide_info,channel_mix_opt))

        button_disabled = len(selected_data)==0
        selected_data_store = json.dumps({'selected_data': selected_data})
        
        return [info_return], [button_disabled], [selected_data_store]

    def disable_channel_selector(self, switched, channel_labels):

        if switched:
            return True, channel_labels
        return False, no_update

    def extract_marker_structures(self, markers_geojson, slide_annotations):

        marked_feature_list = []
        structure_names = [i['properties']['name'] for i in slide_annotations]
        for f in markers_geojson['features']:
            # Getting the info of which structure this marker is marking
            marked_name = f['properties']['name']
            marked_idx = f['properties']['feature_index']

            if marked_name in structure_names:
                marked_feature = slide_annotations[structure_names.index(marked_name)]['features'][marked_idx]
                marked_feature['properties']['name'] = f'Marked {marked_feature["properties"]["name"]}'
                marked_feature_list.append(marked_feature)

        return marked_feature_list

    def download_image_data(self, feature_list:list, x_scale:float, y_scale:float, tile_url:str='', save_masks:bool=False, image_opts:list = [], mask_opts:str = '',save_format:Union[str,list]='PNG', combine:bool=False, save_path:str=''):
        
        # Scaling coordinates of features back to the slide CRS
        if not mask_opts=='Intersecting':
            feature_collection = {
                'type': 'FeatureCollection',
                'features': feature_list
            }
            scaled_features = geojson.utils.map_geometries(lambda g: geojson.utils.map_tuples(lambda c: (c[0]/x_scale,c[1]/y_scale),g),feature_collection)

        else:
            # Creating the intersecting masks
            scaled_feature_list = []
            for geo in feature_list:
                if type(geo)==dict:
                    scaled_feature_list.append(geojson.utils.map_geometries(lambda g: geojson.utils.map_tuples(lambda c: (c[0]/x_scale,c[1]/y_scale),g),deepcopy(geo)))
                elif type(geo)==list:
                    for h in geo:
                        scaled_feature_list.append(geojson.utils.map_geometries(lambda g: geojson.utils.map_tuples(lambda c: (c[0]/x_scale,c[1]/y_scale),g),deepcopy(h)))

            mask_names = [i['properties']['name'] for i in feature_list[1]]

            intersecting_masks = format_intersecting_masks(
                scaled_feature_list[0],
                scaled_feature_list[1:],
                mask_format = 'one-hot-labels'
            )

            scaled_features = scaled_feature_list[0]

        channel_names = [i['name'] for i in image_opts]

        color_opts = None
        if any(['color' in i for i in image_opts]):
            color_opts = [
                [
                    int(i) for i in j['color'].replace('rgba(','').replace(')','').replace(' ','').split(',')[:-1]
                ]
                for j in image_opts
            ]

        if not color_opts is None:
            channel_names = ['red','green','blue']

        for f_idx,f in enumerate(scaled_features['features']):
            if save_masks:
                # This is for a normal tile_url grabbing an RGB image from a non-multi-frame image
                if not any(['frame' in i for i in image_opts]):
                    image, mask = get_feature_image(
                        feature=f,
                        tile_source = tile_url,
                        return_mask = save_masks
                    )
                else:
                    image, mask = get_feature_image(
                        feature = f,
                        tile_source = tile_url,
                        return_mask = save_masks,
                        frame_index = [i['frame'] for i in image_opts],
                        frame_colors=color_opts
                    )

                if combine and save_format=='OME-TIFF':
                    if mask_opts=='Structure Only':
                        combined_image_mask = np.vstack(
                            (
                                np.moveaxis(image,source=-1,destination=0),
                                mask[None,:,:]
                            )
                        )

                        write_ome_tiff(
                            combined_image_mask,
                            save_path+f'/Images & Masks/{f["properties"]["name"]}_{f_idx}.ome.tiff',
                            channel_names+[f['properties']['name']],
                            [1.0,1.0],
                            1.0
                        )
                    else:

                        image = np.moveaxis(image,source=-1,destination = 0)
                        mask = np.moveaxis(intersecting_masks[f_idx],source=-1,destination=0)
                        combined_image_mask = np.vstack(
                            (
                                image,
                                mask
                            )
                        )

                        write_ome_tiff(
                            combined_image_mask,
                            save_path+f'/Images & Masks/{f["properties"]["name"]}_{f_idx}.ome.tiff',
                            channel_names+mask_names,
                            [1.0,1.0],
                            1.0
                        )

                else:
                    img_save_format = save_format
                    mask_save_format = save_format

                    if os.path.exists(f'{save_path}/Images/'):
                        if img_save_format=='OME-TIFF':
                            image = np.moveaxis(image,source=-1,destination=0)

                            write_ome_tiff(
                                image,
                                save_path+f'/Images/{f["properties"]["name"]}_{f_idx}.ome.tiff',
                                channel_names,
                                [1.0,1.0],
                                1.0
                            )
                        elif img_save_format in ['TIFF','PNG','JPG']:
                            if len(np.shape(image))==2 or np.shape(image)[-1]==1 or np.shape(image)[-1]==3 or img_save_format=='TIFF':
                                image_save_path = f'{save_path}/Images/{f["properties"]["name"]}_{f_idx}.{save_format.lower()}'
                                Image.fromarray(image).save(image_save_path)
                            else:
                                image_save_path = f'{save_path}/Images/{f["properties"]["name"]}_{f_idx}.tiff'
                                Image.fromarray(image).save(image_save_path)
                                
                    if os.path.exists(f'{save_path}/Masks/'):
                        if mask_save_format == 'OME-TIFF':
                            if mask_opts=='Structure Only':
                                mask = np.moveaxis(mask,source=-1,destination=0)

                                write_ome_tiff(
                                    mask,
                                    save_path+f'/Masks/{f["properties"]["name"]}_{f_idx}.ome.tiff',
                                    [f["properties"]["name"]],
                                    [1.0,1.0],
                                    1.0
                                )

                            elif mask_opts=='Intersecting':
                                write_ome_tiff(
                                    intersecting_masks[f_idx],
                                    save_path+f'/Masks/{f["properties"]["name"]}_{f_idx}.ome.tiff',
                                    mask_names,
                                    [1.0,1.0],
                                    1.0                                
                                )

                        elif mask_save_format in ['TIFF','PNG','JPG']:
                            if mask_opts=='Structure Only':
                                save_mask = mask
                            elif mask_opts=='Intersecting':
                                save_mask = intersecting_masks[f_idx]

                            if len(np.shape(save_mask))==2 or np.shape(save_mask)[-1]==1 or np.shape(save_mask)[-1]==3 or mask_save_format=='TIFF':
                                # Apply some kind of artificial color if not grayscale or RGB
                                mask_save_path = f'{save_path}/Masks/{f["properties"]["name"]}_{f_idx}.{save_format.lower()}'
                                Image.fromarray(mask).save(mask_save_path)
                            else:
                                # Just overwriting and saving as TIFF anyways
                                mask_save_path = f'{save_path}/Masks/{f["properties"]["name"]}_{f_idx}.tiff'
                                Image.fromarray(mask).save(mask_save_path)

            else:
                
                if any(['frame' in i for i in image_opts]):
                    image = get_feature_image(
                        feature=f,
                        tile_source = tile_url,
                        return_mask = save_masks,
                        frame_index = [i['frame'] for i in image_opts],
                        frame_colors=color_opts
                    )
                else:
                    image = get_feature_image(
                        feature=f,
                        tile_source = tile_url,
                        return_mask = save_masks,
                    )

                img_save_format = save_format
                if img_save_format=='OME-TIFF':
                    write_ome_tiff(
                        image,
                        save_path+f'/Images/{f["properties"]["name"]}_{f_idx}.ome.tiff',
                        channel_names,
                        [1.0,1.0],
                        1.0
                    )
                elif img_save_format in ['TIFF','PNG','JPG']:

                    if len(np.shape(image))==2 or np.shape(image)[-1]==1 or np.shape(image)[-1]==3 or img_save_format=='TIFF':
                        image_save_path = f'{save_path}/Images/{f["properties"]["name"]}_{f_idx}.{save_format.lower()}'
                        Image.fromarray(image).save(image_save_path)
                    else:
                        image_save_path = f'{save_path}/Images/{f["properties"]["name"]}_{f_idx}.tiff'
                        Image.fromarray(image).save(image_save_path)

    def download_property_data(self, feature_list, save_format, save_path):
        
        # Making a dataframe from feature properties:
        property_list = []
        for f in feature_list:
            property_list.append(f['properties'])
        
        structure_name = f['properties']['name']
        property_df = pd.json_normalize(property_list)

        if save_format == 'CSV':
            property_df.to_csv(save_path+f'/{structure_name}_properties.csv')
        elif save_format == 'XLSX':
            with pd.ExcelWriter(save_path+f'/{structure_name}_properties.xlsx') as writer:
                property_df.to_excel(writer,engine='openpyxl')
                writer.close()
            
    def download_annotations(self, feature_list, x_scale, y_scale, save_format, save_path):
        
        feature_collection = {
            'type': 'FeatureCollection',
            'features': feature_list
        }
        structure_name = feature_list[0]['properties']['name']

        feature_collection = geojson.utils.map_geometries(lambda g: geojson.utils.map_tuples(lambda c: (c[0]/x_scale,c[1]/y_scale),g),feature_collection)
        feature_collection['properties'] = {
            'name': structure_name,
            '_id': uuid.uuid4().hex[:24]
        }

        if save_format=='Aperio XML':
            export_format = 'aperio'
            save_path = save_path + f'/{structure_name}.xml'

        elif save_format=='Histomics (JSON)':
            export_format = 'histomics'
            save_path = save_path +f'/{structure_name}.json'

        elif save_format == 'GeoJSON':
            export_format = 'geojson'
            save_path = save_path +f'/{structure_name}.json'
        
        export_annotations(
            feature_collection,
            format = export_format,
            save_path = save_path
        )

    def pick_download_type(self, download_info):

        if download_info['download_type']=='Properties':
            self.download_property_data(
                download_info['features'],
                download_info['format'],
                download_info['folder']
            )
        elif download_info['download_type'] in ['Images','Masks','Images & Masks']:
            self.download_image_data(
                download_info['features'],
                download_info['x_scale'],
                download_info['y_scale'],
                download_info['tile_url'],
                download_info['save_masks'],
                download_info['image_opts'],
                download_info['mask_opts'],
                download_info['format'],
                download_info['combine'],
                download_info['folder']
            )
        elif download_info['download_type'] == 'Annotations':
            self.download_annotations(
                download_info['features'],
                download_info['x_scale'], 
                download_info['y_scale'], 
                download_info['format'],
                download_info['folder']
            )

    def create_zip_file(self, base_path, output_file):
        
        # Writing temporary data to a zip file
        with zipfile.ZipFile(output_file,'w', zipfile.ZIP_DEFLATED) as zip:
            for path,subdirs,files in os.walk(base_path):
                extras_in_path = path.split('/downloads/')[0]+'/downloads/'
                for name in files:
                    if not 'zip' in name:
                        zip.write(os.path.join(path,name),os.path.join(path.replace(extras_in_path,''),name))
        
            zip.close()

    def start_download_data(self, clicked, selected_structures, selected_data, selected_data_formats, slide_annotations, slide_markers, slide_info, channel_mix_frames, channel_mix_colors, channel_mix_checked, selected_data_channels, selected_mask_options):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        interval_disabled = False
        new_values = []
        button_disabled = True
        selected_structures = get_pattern_matching_value(selected_structures)
        selected_data = get_pattern_matching_value(selected_data)
        slide_annotations = json.loads(get_pattern_matching_value(slide_annotations))
        slide_info = json.loads(get_pattern_matching_value(slide_info))
        slide_markers = get_pattern_matching_value(slide_markers)
        
        base_download_folder = uuid.uuid4().hex[:24]
        download_folder = self.assets_folder+'/downloads/'
        if not os.path.exists(download_folder+base_download_folder):
            os.makedirs(download_folder+base_download_folder)

        # Specifying which download threads to deploy
        download_thread_list = []
        layer_names = [i['properties']['name'] for i in slide_annotations]
        for s_idx, struct in enumerate(selected_structures):
            if struct in layer_names:
                struct_features = slide_annotations[layer_names.index(struct)]['features']
            elif struct == 'Marked Structures':
                if not slide_markers is None:
                    struct_features = []
                    for m in slide_markers:
                        struct_features.extend(self.extract_marker_structures(m['props']['data'],slide_annotations))

            for d_idx, (data,data_format,data_channels,data_masks,data_channel_mix) in enumerate(zip(selected_data,selected_data_formats, selected_data_channels, selected_mask_options, channel_mix_checked)):
                if data in ['Images','Masks','Images & Masks']:
                    if not os.path.exists(self.assets_folder+'/downloads/'+base_download_folder+'/'+data):
                        os.makedirs(self.assets_folder+'/downloads/'+base_download_folder+'/'+data)

                    if data_masks=='Intersecting':
                        struct_features = [{'type': 'FeatureCollection', 'features': struct_features}, slide_annotations]

                    if data_channel_mix:
                        image_opts = [
                            {
                                'frame': slide_info['tiles_metadata']['channels'].index(c_name),
                                'name': c_name,
                                'color': channel_mix_colors[c_idx]['color']
                            }
                            for c_idx,c_name in enumerate(channel_mix_frames)
                        ]
                    else:
                        if 'frames' in slide_info['tiles_metadata']:
                            image_opts = [
                                {
                                    'frame': slide_info['tiles_metadata']['channels'].index(c_name),
                                    'name': c_name,
                                }
                                for c_name in data_channels
                            ]
                        else:
                            image_opts = [
                                {
                                    'name': i
                                }
                                for i in ['red','green','blue']
                            ]

                    download_thread_list.append(
                        {
                            'download_type': data,
                            'structure':struct,
                            'x_scale': slide_info['x_scale'],
                            'y_scale': slide_info['y_scale'],
                            'format': data_format,
                            'folder': self.assets_folder+'/downloads/'+base_download_folder,
                            'features': struct_features,
                            'tile_url': slide_info['tiles_url'].replace('zxy/{z}/{x}/{y}','region'),
                            'save_masks': 'Masks' in data,
                            'image_opts': image_opts,
                            'mask_opts': data_masks,
                            'combine': '&' in data,
                            '_id': uuid.uuid4().hex[:24]
                        }
                    )    
                else:
                    download_thread_list.append(
                        {
                            'download_type': data,
                            'structure': struct,
                            'x_scale': slide_info['x_scale'],
                            'y_scale': slide_info['y_scale'],
                            'format': data_format,
                            'folder': self.assets_folder+'/downloads/'+base_download_folder,
                            'features': struct_features,
                            '_id': uuid.uuid4().hex[:24]
                        }
                    )

        download_data_store = json.dumps({
            'selected_data': [],
            'base_folder':self.assets_folder+'/downloads/'+base_download_folder,
            'zip_file_path': self.assets_folder+'/downloads/'+base_download_folder+'/fusion_download.zip',
            'download_tasks': download_thread_list,
            'current_task': download_thread_list[0]['_id'],
            'completed_tasks': []
        })

        new_thread = threading.Thread(
            target = self.pick_download_type,
            name = download_thread_list[0]['_id'],
            args = [download_thread_list[0]],
            daemon = True
        )
        new_thread.start()       

        return [interval_disabled], [new_values], [new_values], [html.Div('Selected data descriptions will appear here')], [button_disabled],[download_data_store]

    def update_download_data(self, new_interval, download_info_store):
        
        new_interval = get_pattern_matching_value(new_interval)
        download_info_store = json.loads(get_pattern_matching_value(download_info_store))

        current_threads = [i.name for i in threading.enumerate()]
        if not 'current_task' in download_info_store:
            raise exceptions.PreventUpdate
        
        if not download_info_store['current_task'] in current_threads:
            download_info_store['completed_tasks'].append(download_info_store['current_task'])
            
            if 'download_tasks' in download_info_store:
                if len(download_info_store['download_tasks'])==1:

                    if not os.path.exists(download_info_store['zip_file_path']):
                        # This means that the last download task was completed, now creating a zip-file of the results
                        zip_files_task = uuid.uuid4().hex[:24]
                        task_name = 'Creating Zip File'
                        download_info_store['current_task'] = zip_files_task
                        del download_info_store['download_tasks'][0]

                        download_progress = 99

                        interval_disabled = False
                        modal_open = True
                        new_n_intervals = no_update
                        download_data = no_update

                        new_thread = threading.Thread(
                            target = self.create_zip_file,
                            name = zip_files_task,
                            args = [download_info_store['base_folder'],download_info_store['zip_file_path']],
                            daemon=True
                        )
                        new_thread.start()
                    else:
                        task_name = 'Zip File Created'

                        download_progress = 99
                        interval_disabled = True
                        modal_open = False
                        new_n_intervals = no_update
                        download_data = dcc.send_file(download_info_store['zip_file_path'])

                elif len(download_info_store['download_tasks'])==0:
                    # This means that the zip file has finished being created
                    task_name = 'All Done!'
                    interval_disabled = False
                    modal_open = False
                    new_n_intervals = 0
                    download_data = dcc.send_file(download_info_store['zip_file_path'])
                    del download_info_store['download_tasks']

                    download_progress = 100

                elif len(download_info_store['download_tasks'])>1:
                    # This means there are still some download tasks remaining, move to the next one
                    interval_disabled = False
                    modal_open = True
                    new_n_intervals = no_update
                    download_data = no_update
                    del download_info_store['download_tasks'][0]

                    download_info_store['current_task'] = download_info_store['download_tasks'][0]['_id']

                    task_name = f'{download_info_store["download_tasks"][0]["structure"]} {download_info_store["download_tasks"][0]["download_type"]}'

                    new_thread = threading.Thread(
                        target = self.pick_download_type,
                        name = download_info_store['current_task'],
                        args = [download_info_store['download_tasks'][0]],
                        daemon=True
                    )
                    new_thread.start()
                    
                    n_complete = len(download_info_store['completed_tasks'])
                    n_remaining = len(download_info_store['download_tasks'])
                    download_progress = int(100*(n_complete/(n_complete+n_remaining)))

                modal_children = html.Div([
                    dbc.ModalHeader(html.H4(f'Download Progress')),
                    dbc.ModalBody([
                        html.Div(
                            html.H6(f'Working on: {task_name}')
                        ),
                        dbc.Progress(
                            value = download_progress,
                            label = f'{download_progress}%'
                        )
                    ])
                ])
            else:
                # Clearing the directory containing download
                download_path = download_info_store['base_folder']
                rmtree(download_path)

                # Continuing with the current download task
                interval_disabled = True
                modal_open = False
                modal_children = no_update
                new_n_intervals = no_update
                download_data = no_update

                del download_info_store['base_folder']
                del download_info_store['zip_file_path']
                del download_info_store['completed_tasks']
                del download_info_store['current_task']

        else:
            # Continuing with the current download task
            interval_disabled = False
            modal_open = True
            modal_children = no_update
            new_n_intervals = no_update
            download_data = no_update
       
        updated_download_info = json.dumps(download_info_store)

        return [updated_download_info],[interval_disabled], [modal_open], [modal_children], [new_n_intervals], [download_data]
        



"""
Base component for FUSION layouts
"""
from typing_extensions import Union
import pandas as pd

# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
from dash import dash_table
from dash_extensions.enrich import (
    DashBlueprint, html, 
    MultiplexerTransform, PrefixIdTransform, BlockingCallbackTransform
)
from dash_extensions.javascript import Namespace

class BaseComponent:
    """
    Base component other components inherit from to initialize basic properties and methods
    """

    title = ''
    description = ''
    component_prefix = None
    database = None
    session_update = None
    assets_folder = None
    js_namespace = None

    def __str__(self):
        return self.title

    def prefix_escape(self,input_id: Union[str,dict])->bool:
        """Specifying the default prefix escape for all blueprints. Basic specification is to ignore "anchor" or ids that already have a prefix

        :param input_id: Input component id either for a layout or a callback
        :type input_id: Union[str,dict]
        :return: True indicates that this id should not receive a prefix transform, False indicates that it should receive a prefix
        :rtype: bool
        """

        if type(input_id)==dict:
            if 'anchor' in input_id['type']:
                return True
            try:
                current_comp_id = int(input_id['type'][0])
                has_id = True
            except ValueError:
                has_id = False
            
            if has_id:
                return True

        elif 'anchor' in input_id:
            return True
        return False
    
    def add_database(self, database:None):
        """Adding a database to the component, connects to the running database for this application.

        :param database: Instance of fusionDB provided by Visualization.get_layout_children()
        :type database: None
        """

        self.database = database
    
    def add_assets_folder(self, assets_folder:str):
        """Adding an assets folder from the Visualization component

        :param assets_folder: String corresponding to assets folder path
        :type assets_folder: str
        """

        self.assets_folder = assets_folder

    def load(self, component_prefix:int):

        self.component_prefix = component_prefix
        self.blueprint = DashBlueprint(
            transforms = [
                PrefixIdTransform(
                    prefix = f'{self.component_prefix}',
                    escape = lambda input_id: self.prefix_escape(input_id)
                ),
                MultiplexerTransform(),
                BlockingCallbackTransform()
            ]
        )

        self.get_callbacks()
        self.get_namespace()
        self.get_clientside_callbacks()

    def get_callbacks(self):
        pass

    def get_clientside_callbacks(self):
        pass

    def get_namespace(self):
        pass

    def gen_layout(self, session_data:dict):

        self.blueprint.layout = self.update_layout(session_data, use_prefix = False)
    
    def update_layout(self, session_data:dict, use_prefix: bool):

        layout = html.Div()

        if use_prefix:
            PrefixIdTransform(prefix = f'{self.component_prefix}').transform_layout(layout)

        return layout

    def make_dash_table(self, df:pd.DataFrame, id: Union[dict,None] = None, editable: bool = False, deletable: bool = False, selectable: bool = True):
        
        return_table = dash_table.DataTable(
            id = id if not id is None else {},
            columns = [{'name':i,'id':i,'deletable':deletable,'selectable':selectable} for i in df],
            data = df.to_dict('records'),
            editable=editable,                                        
            sort_mode='multi',
            sort_action = 'native',
            page_current=0,
            page_size=5,
            style_cell = {
                'overflow':'hidden',
                'textOverflow':'ellipsis',
                'maxWidth':0
            },
            tooltip_data = [
                {
                    column: {'value':str(value),'type':'markdown'}
                    for column, value in row.items()
                } for row in df.to_dict('records')
            ],
            tooltip_duration = None
        )

        return return_table

    def get_user_external_token(self, session_data):
        try:
            user_token = session_data.get('user',{}).get('external',{}).get('token')
        except:
            user_token = None
        return user_token

    def get_user_external_login(self, session_data):
        try:
            user_login = session_data.get('user',{}).get('external',{}).get('login','Guest')
        except:
            user_login = None
        return user_login

    def get_user_external_id(self, session_data):

        try:
            user_id = session_data.get('user',{}).get('external',{}).get('_id')
            if user_id is None:
                user_id = session_data.get('user',{}).get('external',{}).get('id')
        except:
            user_id = None
        return user_id

    def get_user_internal_token(self, session_data):

        try:
            user_token = session_data.get('user',{}).get('token')
        except:
            user_token = None

        return user_token

    def get_user_internal_login(self, session_data):
        
        try:
            user_login = session_data.get('user',{}).get('login')
        except:
            user_login = None

        return user_login

    def get_user_internal_id(self,session_data):
        
        try:
            user_id = session_data.get('user',{}).get('id')
        except:
            user_id = None

        return user_id

    def get_session_id(self, session_data):

        try:
            session_id = session_data.get('session',{}).get('id')
        except:
            session_id = None

        return session_id

    def update_request_str(self, req_str: str, **query_params):

        print(f'{req_str=}')
        for k,v in query_params.items():
            if v is None:
                continue

            if '?' in req_str:
                sep = '&'
            else:
                sep = '?'
            
            req_str += f'{sep}{k}={v}'
        
        return req_str


class BaseSchema:
    """
    Schemas created in FUSION should inherit from this class
    """
    @classmethod
    def from_dict(cls, dict):
        return cls(**dict)


class Tool(BaseComponent):
    """General class for interactive components that visualize, edit, or perform analyses on data.
    """
    session_update = False


class MultiTool(Tool):
    """General class for a Tool which works on multiple slides at once
    """
    session_update = True


#TODO: Anything else that can be bundled with handler?
class Handler:
    pass

class DSATool(MultiTool):
    """A sub-class of Tool specific to DSA components. 
    The only difference is that these components always update relative to the session data.

    :param Tool: General class for components that perform visualization and analysis of data.
    :type Tool: None
    """

    pass


class MapComponent(MultiTool):
    """General class for components added to SlideMap
        For more information see dash-leaflet: https://www.dash-leaflet.com/

    """

    def get_scale_factors(self, image_metadata: dict):
        """Function used to initialize scaling factors applied to GeoJSON annotations to project annotations into the SlideMap CRS (coordinate reference system)

        :return: x and y (horizontal and vertical) scale factors applied to each coordinate in incoming annotations
        :rtype: float
        """
        if not any([i is None for i in [image_metadata.get('sizeX'), image_metadata.get('sizeY'), image_metadata.get('levels')]]):
            base_dims = [
                image_metadata['sizeX']/(2**(image_metadata['levels']-1)),
                image_metadata['sizeY']/(2**(image_metadata['levels']-1))
            ]

            x_scale = base_dims[0] / image_metadata['sizeX']
            y_scale = -(base_dims[1]) / image_metadata['sizeY']


            return x_scale, y_scale
        return 1.0, 1.0








"""Base classes for tools
"""
from typing_extensions import Union
import asyncio


def asyncio_db_loop(method):
    """Decorator for checking that an event loop is present for handling asynchronous callse

    :param method: Function which has asynchronous process
    :type method: None
    """
    def wrapper(self, *args, **kwargs):
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError as e:
            if str(e).startswith('There is no current event loop in thread'):
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
            else:
                raise

        result = method(self, *args, **kwargs)
        return result

    return wrapper








"""

API associated with running application, accessing select elements in fusionDB


"""

import os
import sys

from fastapi import FastAPI, APIRouter, Response, Request, Depends
from fastapi.middleware.cors import CORSMiddleware
import large_image
import requests
import json
from typing import Annotated
from typing_extensions import Union

from copy import deepcopy
import numpy as np
import uvicorn

import asyncio

from shapely.geometry import box, shape
from fusion_tools import asyncio_db_loop
from fusion_tools.database.database import fusionDB 



class fusionAPI:
    def __init__(self,
                database: Union[fusionDB,None],
                cors_options: dict = {}):

        self.database = database
        self.cors_options = cors_options

        # - GET all rows in tables
        # - GET,PUT,POST,OPTIONS by id, each row in tables in models.py

        self.router = APIRouter()

        # User
        self.router.add_api_route('/user', lambda: self.get_from_table("user"), methods=["GET"],tags = ['user'])
        #self.router.add_api_route('/user/me', lambda: self.get_from_table("user"), methods=["GET"],tags = ['user'])
        self.router.add_api_route('/user/authenticate',self.authenticate, methods = ["GET"],tags=['user'])
        self.router.add_api_route('/user/{id}', lambda id: self.get_from_table("user",id), methods=["GET"],tags = ['user'])

        # VisSession 
        self.router.add_api_route('/vis_session', lambda: self.get_from_table("vis_session"), methods=["GET"], tags = ['vis_session'])
        self.router.add_api_route('/vis_session/{id}', lambda id: self.get_from_table("vis_session",id), methods=["GET"], tags = ['vis_session'])

        # Item
        self.router.add_api_route('/item', lambda: self.get_from_table("item"), methods=["GET"],tags=['item'])
        self.router.add_api_route('/item/{id}', lambda id: self.get_from_table("item",id), methods=["GET"],tags=['item'])

        # Layer
        self.router.add_api_route('/layer', lambda: self.get_from_table("layer"), methods=["GET"],tags = ['layer'])
        self.router.add_api_route('/layer/{id}', lambda id: self.get_from_table("layer",id), methods=["GET"],tags = ['layer'])

        # Structure
        self.router.add_api_route('/structure', lambda: self.get_from_table("structure"), methods=["GET"],tags = ['structure'])
        self.router.add_api_route('/structure/{id}', lambda id: self.get_from_table("structure",id), methods=["GET"],tags = ['structure'])

        # ImageOverlay
        self.router.add_api_route('/image_overlay', lambda: self.get_from_table("image_overlay"), methods=["GET"], tags = ['image_overlay'])
        self.router.add_api_route('/image_overlay/{id}', lambda id: self.get_from_table("image_overlay",id), methods=["GET"], tags = ['image_overlay'])

        # Annotation
        self.router.add_api_route('/annotation', lambda: self.get_from_table("annotation"), methods=["GET"], tags = ['annotation'])
        self.router.add_api_route('/annotation/{id}', lambda id: self.get_from_table("annotation",id), methods=["GET"], tags = ['annotation'])

        # Data
        self.router.add_api_route('/data', lambda: self.get_from_table("data"), methods=["GET"], tags = ['data'])
        self.router.add_api_route('/data/{id}', lambda id: self.get_from_table("data",id), methods=["GET"], tags = ['data'])

    @asyncio_db_loop
    def search_db(self, search_kwargs, size, offset):
        
        loop = asyncio.get_event_loop()
        search_output = loop.run_until_complete(
            asyncio.gather(
                self.database.search(
                    search_kwargs = search_kwargs,
                    size = size,
                    offset = offset
                )
            )
        )

        return search_output[0].copy()

    def get_from_table(self, table_name: str, id: str | None = None, request: Request | None = None) -> list:
        """Getting one or more elements from a specific table in the database

        :param table_name: Name of table to GET from
        :type table_name: str
        :param id: ID of specific element in table to GET
        :type id: str | None
        :param request: Query parameters, etc.
        :type request: Request
        """

        token = None
        size = None
        offset = 0
        if not request is None:
            if request.query_params.get('token'):
                token = request.query_params.get('token')
            if request.query_params.get('size'):
                size = request.query_params.get('size')
            if request.query_params.get('offset'):
                offset = request.query_params.get('offset')

        if not token is None:
            user_filter = {
                'user': {
                    'token': token
                }
            }
        else:
            user_filter = {}

        if id is not None:
            id_filter = {
                table_name: {
                    'id': id
                }
            }
        else:
            id_filter = {}

        search_output = self.search_db(
            search_kwargs = {
                'type': table_name,
                'filter': id_filter | user_filter 
            },
            size = size,
            offset = offset
        )

        return search_output

    def authenticate(self, login:str, password:str) -> dict:

        auth_check = self.database.check_user_login_password(login,password)

        return auth_check


